#  Copyright 2021 Google LLC
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from functools import partial
import numpy as np
import jax
from jax.experimental.pjit import pjit
from jax.sharding import PartitionSpec as P
from flax.core.frozen_dict import freeze
from .utils import \
    add_idxes, collate_fn_wrapper, pred_fn_wrapper, default_output_fn


class Predictor:
    def __init__(self,
                 deployer,
                 collate_fn,
                 pred_fn,
                 output_fn=None,
                 params_sharding_rules=None):
        self._deployer = deployer
        self._collate_fn = partial(collate_fn_wrapper, collate_fn=collate_fn)

        self._params_sharding_rules = params_sharding_rules
        self._pred_fn = partial(
            pred_fn_wrapper,
            pred_fn=pred_fn,
            under_pmap=self._deployer.mesh is None)
        self._params_spec = None
        self._p_pred_step = None

        if output_fn is None:
            self._output_fn = default_output_fn
        else:
            self._output_fn = output_fn

    def setup_running_step(self,
                           pred_fn,
                           dummy_batch,
                           params,
                           params_sharding_rules):
        if self._deployer.mesh is None:
            self._p_pred_step = jax.pmap(pred_fn, axis_name='batch')
        else:
            data_spec = jax.tree_util.tree_map(
                lambda x: P(*(('dp',) + (None,) * (len(x.shape) - 1))),
                dummy_batch)

            self._params_spec = self._deployer.get_params_spec(
                params=params, params_sharding_rules=params_sharding_rules)

            self._p_pred_step = pjit(
                pred_fn,
                in_shardings=(None, self._params_spec, data_spec),
                out_shardings=None)

    def predict(self,
                examples,
                per_device_batch_size,
                params,
                params_meshed=False,
                desc=None):
        raw_n_inputs = len(examples)
        _, global_batch_size = self._deployer.process_batch_size(
            per_device_batch_size=per_device_batch_size)
        examples = examples + [examples[0]] * (global_batch_size - 1)
        examples = add_idxes(examples=examples)

        data_batches = self._deployer.get_model_input_batches(
            examples=examples,
            per_device_batch_size=per_device_batch_size,
            collate_fn=self._collate_fn,
            shuffle=False,
            shuffle_rng=None,
            desc=f'Predicting ({desc})' if desc is not None else 'Predicting')

        params = self._deployer.process_to_run_model(freeze(params))

        preds = []
        for batch in data_batches:
            if self._p_pred_step is None:
                self.setup_running_step(
                    pred_fn=self._pred_fn,
                    dummy_batch=batch,
                    params=params,
                    params_sharding_rules=self._params_sharding_rules)

            if (self._params_spec is not None) and (not params_meshed):
                params = self._deployer.shard_params(
                    params=params, params_spec=self._params_spec)
                params_meshed = True

            pred_rng = self._deployer.process_to_run_model(
                self._deployer.gen_rng(), is_prng_key=True)

            batch_preds_with_idxes = self._deployer.run_model_step(
                step_fn=self._p_pred_step,
                input_args=(pred_rng, params, batch))

            batch_preds = self._deployer.process_batch_preds(
                batch_preds_with_idxes=batch_preds_with_idxes)
            batch_preds = jax.tree_util.tree_map(np.asarray, batch_preds)

            batch_preds = self._output_fn(batch_preds)
            assert isinstance(batch_preds, list) and \
                   len(batch_preds) == global_batch_size
            preds.extend(batch_preds)

        return preds[:raw_n_inputs]


#  Copyright 2021 Google LLC
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import json
import os
import re
import gc
import jax
import jax.numpy as jnp
from jax.experimental import multihost_utils
from flax.jax_utils import unreplicate
from flax.core.frozen_dict import freeze, unfreeze
from flax.serialization import (
    msgpack_serialize, msgpack_restore, from_state_dict, to_state_dict)
from flax.traverse_util import flatten_dict, unflatten_dict, empty_node


PARAMS_FILENAME = 'params.msgpack'
PARAMS_INDEX_FILENAME = 'params_index.json'
OPT_STATE_FILENAME = 'opt_state.msgpack'
OPT_STATE_INDEX_FILENAME = 'opt_state_index.json'
RNG_FILENAME = 'rng.npy'


def save_params(mesh, params, ckpt_dir, max_shard_size):
    filepath = f'{ckpt_dir}/{PARAMS_FILENAME}'
    if mesh is not None:
        params = multihost_utils.process_allgather(params)

    if jax.process_index() == 0:
        shards, index = flax_shard_checkpoint(
            params=params, filepath=filepath, max_shard_size=max_shard_size)

        if index is None:
            open(filepath, "wb").write(msgpack_serialize(unfreeze(params)))
        else:
            json.dump(index, open(
                f'{ckpt_dir}/{PARAMS_INDEX_FILENAME}', 'w'), indent=4)
            for shard_file, shard in shards.items():
                open(shard_file, "wb").write(msgpack_serialize(
                    unfreeze(unflatten_dict(shard, sep='/'))))


def save_opt_state(mesh, opt_state, ckpt_dir, max_shard_size):
    filepath = f'{ckpt_dir}/{OPT_STATE_FILENAME}'

    if mesh is None:
        opt_state = unreplicate(opt_state)
    else:
        opt_state = multihost_utils.process_allgather(opt_state)

    if jax.process_index() == 0:
        opt_state = to_state_dict(opt_state)
        shards, index = flax_shard_checkpoint(
            params=opt_state, filepath=filepath, max_shard_size=max_shard_size)

        if index is None:
            open(filepath, "wb").write(msgpack_serialize(unfreeze(opt_state)))
        else:
            json.dump(index, open(
                f'{ckpt_dir}/{OPT_STATE_INDEX_FILENAME}', 'w'), indent=4)
            for shard_file, shard in shards.items():
                open(shard_file, "wb").write(
                    msgpack_serialize(unfreeze(unflatten_dict(shard, sep='/'))))


def save_rng(rng, ckpt_dir):
    if jax.process_index() == 0:
        jnp.save(f'{ckpt_dir}/rng.npy', rng)


def load_params(ckpt_dir):
    filepath = f'{ckpt_dir}/{PARAMS_FILENAME}'
    with jax.default_device(jax.devices('cpu')[0]):
        if os.path.exists(f'{ckpt_dir}/{PARAMS_INDEX_FILENAME}'):
            weight_map = json.load(
                open(f'{ckpt_dir}/{PARAMS_INDEX_FILENAME}'))['weight_map']
            params = load_flax_sharded_weights(
                shard_files=list(set(weight_map.values())))
        else:
            params = msgpack_restore(open(filepath, 'rb').read())

        params = jax.tree_util.tree_map(jnp.asarray, params)

    return params


def load_opt_state(ckpt_dir, params, optimizer):
    filepath = f'{ckpt_dir}/{OPT_STATE_FILENAME}'
    if (not os.path.exists(filepath) and
            not os.path.exists(f'{ckpt_dir}/{OPT_STATE_INDEX_FILENAME}')):
        return None

    with jax.default_device(jax.devices('cpu')[0]):
        if os.path.exists(f'{ckpt_dir}/{OPT_STATE_INDEX_FILENAME}'):
            weight_map = json.load(
                open(f'{ckpt_dir}/{OPT_STATE_INDEX_FILENAME}'))['weight_map']
            opt_state = load_flax_sharded_weights(
                shard_files=sorted(list(set(weight_map.values()))))
        else:
            opt_state = msgpack_restore(open(filepath, 'rb').read())

        opt_state = from_state_dict(
            target=optimizer.init(freeze(params)), state=opt_state)
    return opt_state


def load_rng(ckpt_dir):
    return jnp.load(f'{ckpt_dir}/{RNG_FILENAME}')


def load_flax_sharded_weights(shard_files):
    """
    Copied from https://github.com/huggingface/transformers/blob/main/src/transformers/modeling_flax_utils.py#L460

    This is the same as [`flax.serialization.from_bytes`]
    (https:lax.readthedocs.io/en/latest/_modules/flax/serialization.html#from_bytes) but for a sharded checkpoint.

    This load is performed efficiently: each checkpoint shard is loaded one by one in RAM and deleted after being
    loaded in the model.

    Args:
        shard_files (`List[str]`:
            The list of shard files to load.

    Returns:
        `Dict`: A nested dictionary of the model parameters, in the expected format for flax models : `{'model':
        {'params': {'...'}}}`.
    """

    # Load the index
    state_sharded_dict = {}
    for shard_file in shard_files:
        # load using msgpack utils
        state = msgpack_restore(open(shard_file, 'rb').read())
        state = flatten_dict(state, keep_empty_nodes=True, sep="/")
        state_sharded_dict.update(state)
        del state
        gc.collect()

    # the state dict is unflattened to the match the format of model.params
    return unflatten_dict(state_sharded_dict, sep="/")


def convert_file_size_to_int(size):
    """
    Copied from https://github.com/huggingface/transformers/blob/main/src/transformers/utils/hub.py#L914

    Converts a size expressed as a string with digits an unit (like `"5MB"`) to an integer (in bytes).

    Args:
        size (`int` or `str`): The size to convert. Will be directly returned if an `int`.

    Example:
    ```py
    >>> convert_file_size_to_int("1MiB")
    1048576
    ```
    """
    if isinstance(size, int):
        return size
    if size.upper().endswith("GIB"):
        return int(size[:-3]) * (2**30)
    if size.upper().endswith("MIB"):
        return int(size[:-3]) * (2**20)
    if size.upper().endswith("KIB"):
        return int(size[:-3]) * (2**10)
    if size.upper().endswith("GB"):
        int_size = int(size[:-2]) * (10**9)
        return int_size // 8 if size.endswith("b") else int_size
    if size.upper().endswith("MB"):
        int_size = int(size[:-2]) * (10**6)
        return int_size // 8 if size.endswith("b") else int_size
    if size.upper().endswith("KB"):
        int_size = int(size[:-2]) * (10**3)
        return int_size // 8 if size.endswith("b") else int_size
    raise ValueError("`size` is not in a valid format. Use an integer followed by the unit, e.g., '5GB'.")


def dtype_byte_size(dtype):
    """
    Copied from https://github.com/huggingface/transformers/blob/main/src/transformers/modeling_flax_utils.py#L84

    Returns the size (in bytes) occupied by one parameter of type `dtype`. Example:
    ```py
    >>> dtype_byte_size(np.float32)
    4
    ```
    """
    if dtype == bool:
        return 1 / 8
    bit_search = re.search(r"[^\d](\d+)$", dtype.name)
    if bit_search is None:
        raise ValueError(f"`dtype` is not a valid dtype: {dtype}.")
    bit_size = int(bit_search.groups()[0])
    return bit_size // 8



def flax_shard_checkpoint(params, filepath, max_shard_size="10GB"):
    """
    Copied from https://github.com/huggingface/transformers/blob/main/src/transformers/modeling_flax_utils.py#L101

    Splits a model state dictionary in sub-checkpoints so that the final size of each sub-checkpoint does not exceed a
    given size. The sub-checkpoints are determined by iterating through the `state_dict` in the order of its keys, so
    there is no optimization made to make each sub-checkpoint as close as possible to the maximum size passed. For
    example, if the limit is 10GB and we have weights of sizes [6GB, 6GB, 2GB, 6GB, 2GB, 2GB] they will get sharded as
    [6GB], [6+2GB], [6+2+2GB] and not [6+2+2GB], [6+2GB], [6GB].

    <Tip warning={true}>

    If one of the model's weight is bigger that `max_shard_size`, it will end up in its own sub-checkpoint which will
    have a size greater than `max_shard_size`.

    </Tip>

    Args:
        params (`Union[Dict, FrozenDict]`): A `PyTree` of model parameters.
        max_shard_size (`int` or `str`, *optional*, defaults to `"10GB"`):
            The maximum size of each sub-checkpoint. If expressed as a string, needs to be digits followed by a unit
            (like `"5MB"`).
    """
    max_shard_size = convert_file_size_to_int(max_shard_size)

    sharded_state_dicts = []
    current_block = {}
    current_block_size = 0
    total_size = 0

    # flatten the weights to chunk
    weights = flatten_dict(params, keep_empty_nodes=True, sep="/")
    for item in weights:
        if weights[item] is empty_node:
            current_block[item] = empty_node
            continue

        weight_size = weights[item].size * dtype_byte_size(weights[item].dtype)

        # If this weight is going to tip up over the maximal size, we split.
        if current_block_size + weight_size > max_shard_size:
            sharded_state_dicts.append(current_block)
            current_block = {}
            current_block_size = 0

        current_block[item] = weights[item]
        current_block_size += weight_size
        total_size += weight_size

    # Add the last block
    sharded_state_dicts.append(current_block)

    # If we only have one shard, we return it
    if len(sharded_state_dicts) == 1:
        return {filepath: sharded_state_dicts[0]}, None

    # Otherwise, let's build the index
    weight_map = {}
    shards = {}
    for idx, shard in enumerate(sharded_state_dicts):
        shard_file = filepath.replace(
            ".msgpack",
            f"-{idx+1:05d}-of-{len(sharded_state_dicts):05d}.msgpack")
        shards[shard_file] = shard
        for weight_name in shard.keys():
            weight_map[weight_name] = shard_file

    # Add the metadata
    metadata = {"total_size": total_size}
    index = {"metadata": metadata, "weight_map": weight_map}
    return shards, index




#  Copyright 2021 Google LLC
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from functools import partial
import json
import numpy as np
import jax
from jax.experimental.pjit import pjit
from jax.sharding import PartitionSpec as P
from flax.jax_utils import replicate
from flax.training import train_state
from flax.traverse_util import flatten_dict
from flax.core.frozen_dict import freeze

from .utils import default_train_step, default_eval_step
from ..predictors import Predictor


class Trainer:
    def __init__(self,
                 deployer,
                 collate_fn,
                 apply_fn,
                 loss_fn,
                 params,
                 optimizer,
                 lr_schedule_fn=None,
                 accumulate_grad_batches=None,
                 params_sharding_rules=None):
        self._deployer = deployer
        self._collate_fn = collate_fn
        self._apply_fn = apply_fn
        self._loss_fn = loss_fn
        self._optimizer = optimizer
        self._lr_schedule_fn = lr_schedule_fn
        self._accumulate_grad_batches = accumulate_grad_batches
        self._params_sharding_rules = params_sharding_rules

        self._state = None
        self._state_spec = None
        self._p_train_step = None
        self._p_eval_step = None

        n_params = sum([param.size for param in flatten_dict(params).values()])
        self._deployer.log_info(f'{n_params:,}', title='Training parameters')

        last_ckpt_info_path = \
            f'{self.workdir}/ckpts/last_ckpt_info.json'
        if os.path.exists(last_ckpt_info_path):
            self._last_ckpt_info = json.load(open(last_ckpt_info_path))
            ckpt_name = self._last_ckpt_info['last_ckpt_name']
            ckpt_dir = f'{self.workdir}/ckpts/{ckpt_name}'

            params = self._deployer.load_params(ckpt_dir=ckpt_dir)
            opt_state = self._deployer.load_opt_state(
                ckpt_dir=ckpt_dir, params=params, optimizer=optimizer)

            self.set_train_state(
                apply_fn=self._apply_fn,
                params=params,
                optimizer=self._optimizer,
                step=self._last_ckpt_info['last_step'],
                opt_state=opt_state)

            self._deployer.load_rng(ckpt_dir=ckpt_dir)

            self._deployer.log_info(
                'loaded last_ckpt \"{last_ckpt_name}\",'
                ' last_step={last_step},'
                ' last_epoch_idx={last_epoch_idx}'.format(
                    **self._last_ckpt_info))
        else:
            self._last_ckpt_info = \
                {'last_ckpt_name': None, 'last_step': 0, 'last_epoch_idx': -1}
            self.set_train_state(
                apply_fn=apply_fn, params=params, optimizer=optimizer, step=0)

        self._default_predictor_fn = partial(
            Predictor,
            deployer=deployer,
            collate_fn=collate_fn,
            params_sharding_rules=params_sharding_rules)

    def set_train_state(
            self, apply_fn, params, optimizer, step, opt_state=None):
        params = freeze(params)
        if opt_state is None:
            with jax.default_device(jax.devices('cpu')[0]):
                opt_state = optimizer.init(params)

        if self._deployer.mesh is None:
            self._state = train_state.TrainState(
                step=step,
                apply_fn=apply_fn,
                params=params,
                tx=optimizer,
                opt_state=opt_state)
            self._state = replicate(self._state)
        else:
            params_spec = self._deployer.get_params_spec(
                params=params,
                params_sharding_rules=self._params_sharding_rules)
            opt_state_spec = self._deployer.get_opt_state_spec(
                params=params, params_spec=params_spec, optimizer=optimizer)

            params = self._deployer.shard_params(
                params=params, params_spec=params_spec)
            opt_state = self._deployer.shard_params(
                params=opt_state, params_spec=opt_state_spec)

            self._state = train_state.TrainState(
                apply_fn=apply_fn,
                params=params,
                tx=optimizer,
                opt_state=opt_state,
                step=step)

            self._state_spec = train_state.TrainState(
                apply_fn=apply_fn,
                params=params_spec,
                tx=optimizer,
                opt_state=opt_state_spec,
                step=None)

    def setup_running_step(self, dummy_batch):
        train_step_fn = partial(
            default_train_step,
            loss_fn=self._loss_fn,
            lr_schedule_fn=self._lr_schedule_fn,
            under_pmap=(self._deployer.mesh is None))

        eval_step_fn = partial(
            default_eval_step,
            loss_fn=self._loss_fn,
            under_pmap=(self._deployer.mesh is None))

        if self._deployer.mesh is None:
            self._p_train_step = jax.pmap(train_step_fn, axis_name='batch')
            self._p_eval_step = jax.pmap(eval_step_fn, axis_name='batch')
        else:
            data_spec = jax.tree_util.tree_map(
                lambda x: P(*(('dp',) + (None,) * (len(x.shape) - 1))),
                dummy_batch)

            self._p_train_step = pjit(
                train_step_fn,
                in_shardings=(None, self._state_spec, data_spec),
                out_shardings=(self._state_spec, None),
                donate_argnums=(1, ))

            self._p_eval_step = pjit(
                eval_step_fn,
                in_shardings=(self._state_spec, data_spec),
                out_shardings=None)

    def train(self, examples, per_device_batch_size, desc=None):
        data_batches = self._deployer.get_model_input_batches(
            examples=examples,
            per_device_batch_size=per_device_batch_size,
            collate_fn=self._collate_fn,
            shuffle=True,
            shuffle_rng=self._deployer.gen_rng(),
            desc=f'Training ({desc})' if desc is not None else 'Training',
            is_train=True,
            accumulate_grad_batches=self._accumulate_grad_batches)

        for batch in data_batches:
            if self._p_train_step is None:
                self.setup_running_step(dummy_batch=batch)

            train_rng = self._deployer.process_to_run_model(
                self._deployer.gen_rng(), is_prng_key=True)
            self._state, metrics = self._deployer.run_model_step(
                step_fn=self._p_train_step,
                input_args=(train_rng, self._state, batch))

            metrics = self._deployer.process_to_deliver(metrics)
            data_batches.set_postfix(**metrics)
            self._deployer.log_metrics(metrics=metrics, step=self.step)

    def eval_loss(self, examples, per_device_batch_size, desc=None):
        data_batches = self._deployer.get_model_input_batches(
            examples=examples,
            per_device_batch_size=per_device_batch_size,
            collate_fn=self._collate_fn,
            shuffle=False,
            shuffle_rng=None,
            desc=f'Evaluating ({desc})' if desc is not None else 'Evaluating')

        losses = []
        for batch in data_batches:
            if self._p_eval_step is None:
                self.setup_running_step(dummy_batch=batch)

            metrics = self._deployer.run_model_step(
                step_fn=self._p_eval_step, input_args=(self._state, batch))

            metrics = self._deployer.process_to_deliver(metrics)
            data_batches.set_postfix(**metrics)

            losses.append(metrics['loss'].item())

        return np.mean(losses).item()

    def fit(self,
            train_examples,
            per_device_batch_size,
            n_epochs,
            eval_examples=None,
            eval_per_device_batch_size=None,
            eval_loss=True,
            eval_predictor=None,
            eval_metric_fn=None,
            eval_sanity_check=True,
            save_every_ckpt=False,
            save_last_ckpt=False,
            save_argmin_ckpt_by_metrics=None,
            save_argmax_ckpt_by_metrics=None,
            save_opt_states=True,
            ckpt_max_shard_size='10GB'):
        if eval_per_device_batch_size is None:
            eval_per_device_batch_size = per_device_batch_size

        if save_argmax_ckpt_by_metrics is None:
            save_argmax_ckpt_by_metrics = []
        if save_argmin_ckpt_by_metrics is None:
            save_argmin_ckpt_by_metrics = []
        min_metrics, max_metrics = {}, {}

        if os.path.exists(f'{self.workdir}/min_metrics.json'):
            min_metrics = json.load(open(
                f'{self.workdir}/min_metrics.json'))
            self._deployer.log_info(min_metrics, title='Detected min_metrics')

        if os.path.exists(f'{self.workdir}/max_metrics.json'):
            max_metrics = json.load(open(
                f'{self.workdir}/max_metrics.json'))
            self._deployer.log_info(max_metrics, title='Detected max_metrics')

        if eval_sanity_check and eval_examples is not None:
            rng_backup = self._deployer._rng
            _, eval_global_batch_size = self._deployer.process_batch_size(
                per_device_batch_size=eval_per_device_batch_size)

            if eval_loss:
                self.eval_loss(
                    examples=eval_examples[:eval_global_batch_size],
                    per_device_batch_size=eval_per_device_batch_size,
                    desc=f'Sanity check')
                self._deployer.log_info(
                    'Sanity check (for evaluation loss) passed.')

            if eval_predictor is not None:
                preds = eval_predictor.predict(
                    examples=eval_examples[:eval_global_batch_size],
                    params=self.params,
                    params_meshed=(self._deployer.mesh is not None),
                    per_device_batch_size=eval_per_device_batch_size,
                    desc=f'Sanity check')
                self._deployer.log_info(
                    'Sanity check (for prediction) passed.')

                if eval_metric_fn is not None:
                    json.dumps(eval_metric_fn(
                        examples=eval_examples[:eval_global_batch_size],
                        preds=preds))
                    self._deployer.log_info(
                        'Sanity check (for evaluation metrics) passed.')

            self._deployer._rng = rng_backup

        for epoch_idx in range(
                self._last_ckpt_info['last_epoch_idx'] + 1, n_epochs):
            if isinstance(train_examples, list):
                epoch_train_examples = train_examples
            else:
                epoch_train_examples = train_examples(epoch_idx=epoch_idx)

            self.train(
                examples=epoch_train_examples,
                per_device_batch_size=per_device_batch_size,
                desc=f'epoch {epoch_idx} / {n_epochs}')

            save_ckpt_kwargs = {
                'epoch_idx': epoch_idx,
                'ckpt_dir_prefix': f'{self.workdir}/ckpts',
                'save_opt_state': save_opt_states,
                'max_shard_size': ckpt_max_shard_size
            }

            if eval_examples is None:
                self._deployer.log_info(
                    'No evaluation cuz \'eval_examples\' is None.')
            else:
                eval_metrics = {}

                if eval_loss:
                    loss = self.eval_loss(
                        examples=eval_examples,
                        per_device_batch_size=eval_per_device_batch_size,
                        desc=f'epoch {epoch_idx} / {n_epochs}')
                    eval_metrics['loss'] = loss

                if eval_predictor is not None:
                    preds = eval_predictor.predict(
                        examples=eval_examples,
                        params=self.params,
                        params_meshed=(self._deployer.mesh is not None),
                        per_device_batch_size=eval_per_device_batch_size,
                        desc=f'epoch {epoch_idx} / {n_epochs}')

                    if eval_metric_fn is not None:
                        eval_metrics.update(eval_metric_fn(
                            examples=eval_examples, preds=preds))

                    eval_outputs = [
                        {'example': example, 'pred': pred}
                        for example, pred in zip(eval_examples, preds)]

                    self._deployer.save_outputs(
                        outputs=eval_outputs,
                        desc=f'epoch{epoch_idx}',
                        step=self.step)

                self._deployer.log_info(
                    info=json.dumps(eval_metrics, indent=4),
                    title=f'Eval results',
                    step=self.step)
                self._deployer.log_metrics(metrics={
                    f'eval_{key}': value
                    for key, value in eval_metrics.items()
                }, step=self.step)

                if self.workdir is not None:
                    result_filepath = \
                        f'{self.workdir}/eval_results_epoch{epoch_idx}.json'
                    json.dump(
                        eval_metrics, open(result_filepath, 'w'), indent=4)
                    self._deployer.log_info(
                        f'eval_results saved into {result_filepath}.')

                for key in save_argmin_ckpt_by_metrics:
                    assert self.workdir is not None
                    if eval_metrics[key] < min_metrics.get(key, float('inf')):
                        min_metrics[key] = eval_metrics[key]

                        if jax.process_index() == 0:
                            self._deployer.log_info(
                                f'minimal {key} updated to {min_metrics[key]}')
                            json.dump(max_metrics, open(
                                f'{self.workdir}/max_metrics.json', 'w'))

                        self.save_ckpt(
                            ckpt_name=f'min_{key}', **save_ckpt_kwargs)

                for key in save_argmax_ckpt_by_metrics:
                    assert self.workdir is not None
                    if eval_metrics[key] > max_metrics.get(key, float('-inf')):
                        max_metrics[key] = eval_metrics[key]

                        if jax.process_index() == 0:
                            self._deployer.log_info(
                                f'maximal {key} updated to {max_metrics[key]}')
                            json.dump(max_metrics, open(
                                f'{self.workdir}/max_metrics.json', 'w'))

                        self.save_ckpt(
                            ckpt_name=f'max_{key}', **save_ckpt_kwargs)

            if save_every_ckpt:
                self.save_ckpt(
                    ckpt_name=f'epoch_{epoch_idx}', **save_ckpt_kwargs)
            elif save_last_ckpt:
                self.save_ckpt(ckpt_name='last', **save_ckpt_kwargs)

    def save_ckpt(self,
                  epoch_idx,
                  ckpt_dir_prefix,
                  ckpt_name,
                  save_opt_state,
                  max_shard_size):
        ckpt_dir = f'{ckpt_dir_prefix}/{ckpt_name}'
        if jax.process_index() == 0:
            os.makedirs(ckpt_dir, exist_ok=True)

        self._deployer.save_params(
            params=self.params,
            ckpt_dir=ckpt_dir,
            max_shard_size=max_shard_size)
        self._deployer.save_rng(ckpt_dir=ckpt_dir)

        if save_opt_state:
            self._deployer.save_opt_state(
                opt_state=self._state.opt_state,
                ckpt_dir=ckpt_dir,
                max_shard_size=max_shard_size)

        if jax.process_index() == 0:
            last_ckpt_info = {
                'last_ckpt_name': ckpt_name,
                'last_step': self.step,
                'last_epoch_idx': epoch_idx
            }
            json.dump(last_ckpt_info, open(
                f'{ckpt_dir_prefix}/last_ckpt_info.json', 'w'), indent=4)
            self._deployer.log_info(
                f'{ckpt_dir_prefix}/last_ckpt_info.json updated.')

    def get_default_predictor(self, pred_fn, output_fn=None):
        return self._default_predictor_fn(pred_fn=pred_fn, output_fn=output_fn)

    @property
    def params(self):
        return self._deployer.process_to_deliver(self._state.params)

    @property
    def step(self):
        return self._deployer.process_to_deliver(self._state.step).item()

    @property
    def workdir(self):
        return self._deployer.workdir
